;;; -*- Mode: Emacs-Lisp -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  File:    pig.el
;;;  Desc.:   PIG mode, mode to process Chinese Pinyin Input to Guobiao
;;;  Version: 1.1
;;;  Author:  Ping Zhou  zhou@okra.fnal.gov
;;;  Date:    November 16, 1992
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Acknowledgment:
;;; Part of the code are grabbed from ispell and hz2gb package with no or 
;;; minimal change. PinYin table was converted from one file by Ed Lai.

;; This program is free software.
;; You can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; See readme file for installation instruction and usage information.
;;;
;;; Bug report:
;;; If you find any bugs, or have any comments, suggestions on the program, 
;;; or made improvements to the code, please send them to the author at
;;; zhou@okra.fnal.gov
;;; Thanks in advance.


(provide 'pig-mode)

(defvar Pig-in-Mule nil
 "Non-nil means you are using pig-mode in Mule.")

(defvar Pig-in-Emacs t
 "Non-nile means you using pig-mode in Emacs. Default to t.")

(defvar Pig-8bit-capable nil
 "*Non-nil means you are using a 8-bit capable Emacs.")
(if (boundp 'mule-version)
  (setq Pig-in-Mule t
	Pig-in-Emacs nil
	Pig-8bit-capable t))

(defvar Pig-PY-table-file "~/zw/ZiDian.gb")
(defvar Pig-ZiDian "~/zw/ZiDian.el"
 "*ZiDian file for PinYin Input to GuoBiao.
   This is generated by the pig-build-ZiDian command.
   Change it if you don't like the name or directory")

(defvar Pig-font-file "~/lib/gb/Fsong-16.cclib")

(defvar Pig-CiDian "~/zw/CiDian.el"
 "*CiDian file for PinYin Input to GuoBiao.")

(defvar Pig-CiDian-file "~/zw/CiDian.gb"
 "*CiDian definition file for PinYin Input to GuoBiao.")

(defvar Pig-CD-hashed nil
 "Working CiDian used internally.")
(setq Pig-CD-hashed nil)		;temporary

(defvar Pig-CD-chang 0
 "Number of hashed entries in Pig-CiDian.")
(setq Pig-CD-chang 0)			;temporary

(defvar Pig-PY-string
  (concat "^\\("
	  "bang\\|beng\\|bing\\|cang\\|ceng\\|cong\\|chang\\|cheng\\|"
	  "chong\\|chuang\\|dang\\|ang\\|deng\\|ding\\|dong\\|fang\\|"
	  "feng\\|gang\\|geng\\|gong\\|guang\\|hang\\|heng\\|hng\\|"
	  "hong\\|huang\\|jiang\\|jing\\|jiong\\|kang\\|keng\\|kong\\|"
	  "kuang\\|lang\\|leng\\|liang\\|ling\\|long\\|mang\\|meng\\|"
	  "ming\\|nang\\|neng\\|niang\\|ning\\|nong\\|pang\\|peng\\|"
	  "ping\\|qiang\\|qing\\|qiong\\|rang\\|reng\\|rong\\|sang\\|"
	  "seng\\|shang\\|sheng\\|shuang\\|song\\|tang\\|teng\\|ting\\|"
	  "tong\\|wang\\|weng\\|xiang\\|xing\\|xiong\\|yang\\|ying\\|"
	  "yong\\|zang\\|zeng\\|zong\\|zhang\\|zheng\\|zhong\\|zhuang\\|"
	  "ng\\|bai\\|cai\\|chai\\|chuai\\|dai\\|gai\\|guai\\|hai\\|"
	  "huai\\|kai\\|kuai\\|lai\\|mai\\|nai\\|pai\\|sai\\|shai\\|"
	  "shuai\\|tai\\|wai\\|zai\\|zhai\\|zhuai\\|ai\\|ban\\|bian\\|"
	  "can\\|chan\\|chuan\\|cuan\\|dan\\|dian\\|duan\\|fan\\|gan\\|"
	  "guan\\|han\\|huan\\|jian\\|juan\\|kan\\|kuan\\|lan\\|lian\\|"
	  "luan\\|man\\|mian\\|nan\\|nian\\|nuan\\|pan\\|pian\\|qian\\|"
	  "quan\\|ran\\|ruan\\|san\\|shan\\|shuan\\|suan\\|tan\\|tian\\|"
	  "tuan\\|wan\\|xian\\|xuan\\|yan\\|yuan\\|zan\\|zhan\\|zhuan\\|"
	  "zuan\\|an\\|bao\\|biao\\|cao\\|chao\\|dao\\|diao\\|gao\\|"
	  "hao\\|jiao\\|kao\\|lao\\|liao\\|mao\\|miao\\|nao\\|niao\\|"
	  "pao\\|piao\\|qiao\\|rao\\|sao\\|shao\\|tao\\|tiao\\|xiao\\|"
	  "yao\\|zao\\|zhao\\|ao\\|bei\\|dei\\|fei\\|gei\\|hei\\|lei\\|"
	  "mei\\|nei\\|pei\\|shei\\|tei\\|wei\\|zei\\|zhei\\|ei\\|ben\\|"
	  "cen\\|chen\\|fen\\|gen\\|hen\\|ken\\|men\\|nen\\|pen\\|ren\\|"
	  "sen\\|shen\\|wen\\|zen\\|zhen\\|en\\|dia\\|jia\\|lia\\|qia\\|"
	  "xia\\|bie\\|die\\|jie\\|lie\\|mie\\|nie\\|pie\\|qie\\|tie\\|"
	  "xie\\|bin\\|jin\\|lin\\|min\\|nin\\|pin\\|qin\\|xin\\|yin\\|"
	  "diu\\|jiu\\|liu\\|miu\\|niu\\|qiu\\|xiu\\|chou\\|cou\\|dou\\|"
	  "fou\\|gou\\|hou\\|kou\\|lou\\|mou\\|nou\\|pou\\|rou\\|shou\\|"
	  "sou\\|tou\\|you\\|zhou\\|zou\\|ou\\|gua\\|hua\\|kua\\|shua\\|"
	  "zhua\\|jue\\|lue\\|nue\\|que\\|xue\\|yue\\|chui\\|cui\\|"
	  "dui\\|gui\\|hui\\|kui\\|rui\\|shui\\|sui\\|tui\\|zhui\\|"
	  "zui\\|chun\\|cun\\|dun\\|gun\\|hun\\|jun\\|kun\\|lun\\|qun\\|"
	  "run\\|shun\\|sun\\|tun\\|xun\\|yun\\|zhun\\|zun\\|chuo\\|"
	  "cuo\\|duo\\|guo\\|huo\\|kuo\\|luo\\|nuo\\|ruo\\|shuo\\|suo\\|"
	  "tuo\\|zhuo\\|zuo\\|ba\\|ca\\|cha\\|da\\|fa\\|ga\\|ha\\|"
	  "ka\\|la\\|ma\\|na\\|pa\\|sa\\|sha\\|ta\\|wa\\|ya\\|za\\|"
	  "zha\\|a\\|ce\\|che\\|de\\|er\\|ge\\|he\\|ke\\|le\\|me\\|ne\\|"
	  "re\\|se\\|she\\|te\\|ye\\|ze\\|zhe\\|e\\|bi\\|chi\\|ci\\|di\\|"
	  "ji\\|li\\|mi\\|ni\\|pi\\|qi\\|ri\\|shi\\|si\\|ti\\|xi\\|yi\\|"
	  "zhi\\|zi\\|bo\\|fo\\|lo\\|mo\\|po\\|wo\\|wo\\|yo\\|bu\\|"
	  "chu\\|cu\\|du\\|fu\\|gu\\|hu\\|ju\\|ku\\|lu\\|mu\\|nu\\|pu\\|"
	  "qu\\|ru\\|shu\\|su\\|tu\\|wu\\|xu\\|yu\\|zhu\\|zu\\|"
	  "lv\\|nv\\|m\\|n"
	  "\\)\\>")
  "All legal spellings of PinYin words. Order is very important here.")

(defvar Pig-SM-string
 (concat
  "^\\(b\\|ch\\|c\\|d\\|f\\|g\\|h\\|j\\|k\\|l\\|m\\|n\\|p\\|q\\|"
  "r\\|sh\\|s\\|t\\|w\\|x\\|y\\|zh\\|z\\)")
 "*ShengMu string for Chinese PinYin.")

(defvar Pig-YM-string 
 (concat
  "\\(iang\\|iong\\|uang\\|"
  "ang\\|eng\\|ian\\|iao\\|iou\\|ing\\|ong\\|uai\\|uan\\|uei\\|uen\\|van"
  "\\|ai\\|an\\|ao\\|ei\\|en\\|ia\\|ie\\|in\\|iu\\|ou"
  "\\|ua\\|ue\\|ui\\|un\\|uo\\|"
  "a\\|e\\|i\\|o\\|u\\|v\\)\\>")	;longer first
 "*YunMu string for Chinese PinYin.")

(defvar Pig-Symbol-string
      "^\\(douh\\|juh\\|maoh\\|wenh\\|fenh\\|dunh\\|ganth\\|jinth\\)\\>")

(defvar Pig-internal-prefix nil
 "Prefix string needed for displaying HanZi.")

(if (not Pig-in-Mule)
  (defvar Pig-GB-syntax-table
   [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
      0 0 0 0 0 0 0 0 0 0 0 0 0 2 2 0  2 2 2 2 2 2 0 0 0 0 0 0 0 0 0 0
      0 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 0 0 0 0 0
      0 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 0 0 0 0 0
      0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
      0 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
      2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
      2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 0])
 (defvar Pig-GB-syntax-table (make-syntax-table))
 (modify-syntax-entry ?- "w" Pig-GB-syntax-table)
 (modify-syntax-entry ?. "w" Pig-GB-syntax-table))


(defvar Pig-mode-syntax-table
 [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
  0 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2  2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 0
  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
 "Syntax table used in pinyin mode to find word boundaries.")

(defvar Pig-mode-map nil "")
(if Pig-mode-map
  ()
 (setq Pig-mode-map (make-sparse-keymap))
 (define-key Pig-mode-map "\e\C-I" 'pig-word)
 (define-key Pig-mode-map "\e\$" 'pig-region)
 (define-key Pig-mode-map "\C-ci" 'pig-instant-mode)
 (define-key Pig-mode-map "\C-cb" 'pig-batch-mode)
 (define-key Pig-mode-map "\C-ch" 'gb2hz-buffer)
 (define-key Pig-mode-map "\C-cg" 'hz2gb-buffer)
 (define-key Pig-mode-map "\C-c;" 'pig-fen-hao)
 (define-key Pig-mode-map "\C-c." 'pig-ju-hao)
 (define-key Pig-mode-map "\C-c?" 'pig-wen-hao)
 (define-key Pig-mode-map "\C-c!" 'pig-gantan-hao)
 (define-key Pig-mode-map "\C-c:" 'pig-mao-hao)
 (define-key Pig-mode-map "\C-c," 'pig-dou-hao)
 (define-key Pig-mode-map "\C-c`" 'pig-dun-hao)
 (define-key Pig-mode-map "\C-c~" 'pig-tilte)
 (define-key Pig-mode-map "\C-c<" 'pig-left-bracket)
 (define-key Pig-mode-map "\C-c>" 'pig-right-bracket))

(defun pig-mode-initialization ()
  "*Setting up the Pin1Yin1 mode."
  (interactive)
  (load Pig-ZiDian)			;define pinyin table
  (if Pig-CiDian (load Pig-CiDian))
  (if Pig-8bit-capable () (pig-build-GB-bitmap))) ; only for 7bit Emacs

(defun pig-buffer ()
 "*Convert all the Chinese PinYin words in the buffer to GuoBiao and display
   them."
 (interactive)
 (pig-region (point-min) (point-max)))

(defun pig-replace-char (from-str to-char string)
 "Replace all FROM-STR with TO-CHAR in the STRING. STRING will be changed."
 (let ((next 0))
  (while (setq next (string-match from-str string next))
   (aset string next to-char))))

(defun pig-py-parser (word)
 "parse WORD as a PinYin CiZu or a PinYin Zi. Return a list. The first element
of it is the PinYin WORD it self. The 2nd element is a list matched GBs.
The 3rd element is a indicator: t for CiZu, nil for Zi. The 4th element is the Zi list of the CiZu or nil."
 (let ((last 0)
       (chang (length word))
       (next 0)
       py-list cizup)
  (pig-replace-char "5" ?0 word)
  (if (setq next (string-match "[.-]" word))	; a CiZu for sure
    (progn
     (while next
      (setq py-list (append py-list 
			    (list (downcase (substring word last next ))))
	    last (1+ next)
	    next (string-match "[.-]" word last)))
     (setq last (downcase (substring word (match-end 0))))
     (if (string-equal last "") ()
      (setq py-list (append py-list (list last))))
     (setq gb-list (pig-CD-lookup py-list)
	   cizup t)) ; a CiZu
   (if (setq next (string-match "[0-5]" word))
     (progn
      (if (= next (1- chang))
	(setq gb-list (pig-ZD-lookup word))
       (while next		; separated by tones
	(setq next (1+ next)
	      py-list (append py-list 
			      (list (downcase (substring word last next ))))
	      last next
	      next (string-match "[0-5]" word last)))
       (if (= last chang) ()
	(setq py-list (append py-list	; does not end with a tone
			      (list (downcase (substring word last))))))
       (setq gb-list (pig-CD-lookup py-list)
	     cizup t)))
    (setq py-list (mapcar '(lambda (char) ; pure letters
			    (append py-list (char-to-string char)))
			  (downcase word)))
    (setq gb-list (pig-CD-lookup py-list)
	  cizup t)
    (if gb-list ()
     (setq gb-list (pig-ZD-lookup word)
	   cizup nil))))
  (list word gb-list cizup py-list)))

(defun pig-ZD-lookup (word)
 "Get gb for a PinYin WORD. If with no tone, then it's for all possible tones."
 (let ((pinyin word)
       gb-list gb-array tone)
  (if (setq tone (string-match "[0-5]" word))
    (setq pinyin (substring word 0 tone)
	  tone (% (- (aref word tone) 44) 5)))
  (if (setq gb-list (cdr (assoc pinyin Pig-Symbol-table))) ;Symbols
    ()
   (setq gb-array (cdr (assoc pinyin Pig-PinYin-table)))
   (if (null gb-array) (setq gb-list nil)
    (if (integerp tone) 
      (setq gb-list (aref gb-array tone))
     (setq gb-list (aref gb-array 0)
	   gb-list (pig-list-union gb-list (aref gb-array 1))
	   gb-list (pig-list-union gb-list (aref gb-array 2))
	   gb-list (pig-list-union gb-list (aref gb-array 3))
	   gb-list (pig-list-union gb-list (aref gb-array 4))))))
  (if Pig-8bit-capable gb-list
   (mapcar 'pig-gb82d gb-list))))

(defun pig-CD-lookup (cizu)
 "Given CIZU, find and return the possible match in CiDian. Return nil if no
match was found."
 (let* ((ci-chang (length cizu))
	(hash-entry (pig-hash-table-lkup (pig-cizu-hash cizu)
					 Pig-CD-hashed 0 (1- Pig-CD-chang)))
	hlist found gb-list length)
  (if (nth 0 hash-entry)
    (progn
     (setq hlist (nth 2 hash-entry)
	   length (length hlist)
	   found (mapcar 'pig-match-cizu hlist))))
  (while found
   (setq gb-list (pig-list-union gb-list (car found))
	 found (cdr found)))
  (if (or Pig-8bit-capable (null gb-list)) gb-list
   (mapcar 'pig-gb82d gb-list))))

(defun pig-match-cizu (hlist)
 "Match a formatted cizu with parsed-cizu and return the matching gb-list."
 (let ((phrase (car hlist))
       (indx 0))
  (if (catch 'match
       (while (< indx ci-chang)
	(setq zi (nth indx cizu))
	(if (> (length zi) 1)
	  (if (null (string-match (concat "^" zi) (nth indx phrase)))
	    (throw 'match nil)))
	(setq indx (1+ indx)))
       t)				; an all initial cizu
    (nth 1 hlist))))

(defun pig-hash-table-lkup (hash CiDian head tail)
 "Match a HASH number in the CiDian between the element number HEAD and TAIL. 
CiDian is a multi-layer nested list.
Returns a list: (status sequence-number list-of-phrases)
status = nil if no match was found, and s-n is that the match would follow
       = non-nil if it is a real match."
 (let ((indx tail)
       (result (list nil -1 ))
       cell mid)
  (if (< (- tail head) 10)		; small eough range, search one by one
    (progn
     (catch 'done
      (while (>= indx head)
       (setq cell (nth indx CiDian))
       (if (> (car cell) hash)
	 (setq indx (1- indx))		; not yet, keep going
	(if (= (car cell) hash)
	  (progn
	   (setq result (list t indx (cdr cell))) ; match found
	   (throw 'done t))
	 (setq result (list nil indx))	; over, no match
	 (throw 'done t)))))
     result)
   (setq mid (1- (/ (+ head tail) 2))	; binary search
	 cell (nth mid CiDian))
   (if (< hash (car cell))		; in the first half
     (pig-hash-table-lkup hash CiDian head (1- mid))
    (if (> hash (car cell))		; in the last half
      (pig-hash-table-lkup hash CiDian mid tail)
     (list t mid (cdr cell)))))))	; exactly equal. match found

(defun pig-build-CiDian ()
 "*Build a CiDian from a file of phrases and their GBs.
Format: Each Ci consists of PinYin with tone separated with -, space, GBs"
 (interactive)
 (if (null (yes-or-no-p "This will destroy the CiDian in use now. Continue ?"))
   ()
  (setq Pig-CD-hashed nil
	Pig-CD-chang 0)
  (let (buf-end end cizu hash gb-code gb-list entry)
   (unwind-protect
     (save-excursion
      (set-buffer (get-buffer-create "PIG-build-buffer"))
      (erase-buffer)
      (set-syntax-table Pig-GB-syntax-table)
      (insert-file Pig-CiDian-file)
      (setq buf-end (point-max))
      (goto-char (point-min))
      (message "Building CiDian out of %s ..." Pig-CiDian-file)
      (while (< (point) buf-end)
       (if (looking-at "\\w+") ; start with a letter - CiZu
	 (progn
	  (setq end (match-end 0)
		cizu (pig-parse-cz (downcase (buffer-substring (point) end)))
		gb-list nil)
	  (goto-char end)
	  (while (looking-at "[\ \t]*\\(\\w+\\)")
	   (setq end (match-end 1)
		 gb-code (buffer-substring (match-beginning 1) end)
		 gb-list (append gb-list (cons gb-code nil)))
	   (goto-char end))
	  (setq entry (list cizu gb-list))
	  (pig-insert-cizu entry)
	  (forward-line 1))
	(forward-line 1)))
      (pig-save-CiDian))		; skip lines with unknown format
    (kill-buffer "PIG-build-buffer")))))

(defun pig-insert-cizu (entry)
 "Find the correct place in and insert the new ENTRY into CiDian."
 (let* ((cizu (car entry))
	(hash (pig-cizu-hash cizu))
	seqn tpr num)
  (if (null Pig-CD-hashed)		; fresh start
    (setq Pig-CD-hashed (list (list hash entry))
	  Pig-CD-chang (1+ Pig-CD-chang))
   (setq tpr (pig-hash-table-lkup hash Pig-CD-hashed 0 (1- Pig-CD-chang))
	 seqn (nth 1 tpr))
   (if (null (car tpr))		; new entry into CiDian
     (setq Pig-CD-hashed 
	   (pig-list-insert-element Pig-CD-hashed seqn (list hash entry))
	   Pig-CD-chang (1+ Pig-CD-chang))
    (setq tpr (cdr (nth seqn Pig-CD-hashed)) ; append to existing element
	  num (1- (length tpr)))
    (if (equal "" (mapconcat 'pig-match-n-append tpr ""))
      (setcdr (nthcdr num tpr) (list entry)))))))

(defun pig-list-insert-element (cd-list seq entry)
 "Insert into CD-LIST after its SEG-th place a new element ENTRY, return the 
new list."
 (if (< seq 0)				; insert at the beginning
   (append (list entry) cd-list)
  (let ((break (nthcdr seq cd-list)) tail)
   (setq tail (append (list entry) (cdr break)))
   (setcdr break nil)
   (append cd-list tail))))

(defun pig-match-n-append (plist)
 "If the first element of PLIST is the same as cizu, then insert GB part of
entry into the send element of PLIST."
 (if (equal cizu (car plist))
   (progn
    (setcdr plist (list (pig-list-union (nth 1 plist) (nth 1 entry))))
    "1")
  ""))

(defun pig-cizu-hash (i-list)
 "Calculate the identification number for a CiZu with its PinYins in a 
list I_LIST, only the initals are used. Here is the function:
The sign bit is not used. The next 3bits for number of chacracters
The next 20bits store the 4 initial letters from the I_LIST."
 (let ((ci-chang (length i-list))
       (indx 0)
       (shift 0)
       upto sum)
  (setq upto (min 4 ci-chang)
	sum (lsh ci-chang 20))
  (while (< indx upto)
   (setq sum (+ sum (lsh (- (aref (nth indx i-list) 0) 97) shift))
	 shift (+ 5 shift)
	 indx (1+ indx)))
  sum))

(defun pig-hex2gb (two-byte-hex)
 "Translate two-byte hex string to 8bit GB string."
 (let ((hex (downcase two-byte-hex)))
  (concat (char-to-string 
	   (+ (* 16 (pig-hex2int (aref hex 0))) (pig-hex2int (aref hex 1))))
	  (char-to-string
	   (+ (* 16 (pig-hex2int (aref hex 2))) (pig-hex2int (aref hex 3)))))))

(defun pig-hex2int (hex)
 "Convert a HEX (lower case) to an int."
 (if (> hex ?9) (- hex 87) (- hex ?0)))

(defun pig-list-union (lista listb)
 "Append LISTB to LISTA and return the union, with euqal elements removed.
Elements within LISTA or LISTB are not checked with equal."
 (let (cell)
  (while listb
   (setq cell (car listb)
	 listb (cdr listb))
   (if (pig-memequal cell lista) ()
    (setq lista (append lista (cons cell nil)))))
  lista))

(defun pig-memequal (cell alist)
 "Return t if CELL is equal to one of the element of ALIST."
 (catch 'yes
  (let
    ((lista alist))
   (while lista 
    (if (equal cell (car lista)) (throw 'yes t))
    (setq lista (cdr lista))))))

(defun pig-build-ZiDian ()
 "*Read and build ZiDian for translating PinYin to GuoBiao."
 (interactive)
 (if (null (yes-or-no-p "This will destroy the ZiDian in use now. Continue ?"))
   ()
  (setq Pig-PinYin-table nil)
  (let
    ((name "PY-table-buffer")
     maxpoint gb-list gb-code ptr tone list-tone end)
   (unwind-protect
     (save-excursion
      (set-buffer (get-buffer-create name))
      (erase-buffer)
      (set-syntax-table Pig-GB-syntax-table)
      (insert-file-contents Pig-PY-table-file)
      (setq maxpoint (1- (point-max)))
      (message "Building ZiDian out of %s" Pig-PY-table-file)
      (while (< (point) maxpoint)
       (if (looking-at "[a-z]+\\|[A-Z]+") ; start with a letter - PinYin line
	 (progn
	  (setq end (match-end 0)
		word (downcase (buffer-substring (point) end)))
	  (if (string-match Pig-PY-string word)
	    (setq Pig-PinYin-table 
		  (append Pig-PinYin-table
			  (list (cons word (make-vector 5 nil))))
		  ptr (cdr (assoc word Pig-PinYin-table)))
	   (error "Illegal PinYin entry in %s." Pig-PY-table-file))
	  (forward-line 1))	;next line please
;;; Not a PinYin line, may be a tone and GB line, or a continuation of GB line
	(if (looking-at "^ *\\([1-5]\\)") ;tone and GB line
	  (setq tone (1- (string-to-int 
			  (buffer-substring (match-beginning 1) 
					    (goto-char (match-end 1)))))
		gb-list nil))
	(while (looking-at " *\\(\\w+\\)") ; the rest of line
	 (setq end (match-end 1)
	       gb-code (buffer-substring (match-beginning 1) end)
	       gb-list (append gb-list (cons gb-code nil)))
	 (goto-char end))
	(setq list-tone (aref ptr tone))
	(aset ptr tone (append list-tone gb-list))
	(forward-line 1)))
      (erase-buffer)
      (insert "(setq Pig-Symbol-table \'"
	      (prin1-to-string Pig-Symbol-table) "\)\n")
      (insert "(setq Pig-PinYin-table \'" 
	      (prin1-to-string Pig-PinYin-table) "\)\n")
      (if (setq name (expand-file-name
		      (read-file-name "Save the new ZiDian to:" Pig-ZiDian)))
	(progn
	 (write-file name)
	 (setq name (buffer-name)))))
    (kill-buffer name)))))

(defun pig-extract-ZiDian ()
 "*Build GB ZiDian."
 (interactive)
 (load "forms" nil t)
 (let
   ((name "PY-table-buffer")
    (ZD-tmp Pig-PinYin-table)
    (ZD-chang (length Pig-PinYin-table))
    maxpoint gb-list gb-code ptr tone list-tone py-kind end)
  (save-excursion
   (set-buffer (get-buffer-create name))
   (erase-buffer)
   (message "Extracting current ZiDian ...")
   (while ZD-tmp
    (setq entry (car ZD-tmp))
    (insert (car entry) "\n")
    (setq tone 1)
    (while (< tone 6)
     (if (setq tone-list (aref tone (cdr entry)))
       (progn
	(insert " " (int-to-string tone) " ")
	(while tone-list
	 (insert (car tone-list))
	 (setq tone-list (cdr tone-list)))
	(insert "\n")))
     (setq tone (1+ tone)))
    (setq ZD-tmp (cdr ZD-tmp)))
   (if (setq name (expand-file-name
		   (read-file-name "Save the GB ZiDian table to:"
				   Pig-ZiDian-file)))
     (progn
      (write-file name)
      (kill-buffer (buffer-name)))))))

(defun pig-save-CiDian ()
 "*Save the CiDian to a new .el file."
 (interactive)
 (let ((name "Pig-CiDian-tmp"))
  (unwind-protect
    (save-excursion
     (set-buffer (get-buffer-create name))
     (if Pig-in-Mule (set-file-coding-system *euc-china*))
     (insert "(setq Pig-CD-chang " (prin1-to-string Pig-CD-chang) "\)\n")
     (insert "(setq Pig-CD-hashed \'"
	     (prin1-to-string Pig-CD-hashed) "\)\n")
     (if (setq name (expand-file-name
		     (read-file-name "Save the new CiDian to:" Pig-CiDian)))
       (progn
	(write-file name)
	(setq name (buffer-name)))))
   (kill-buffer name))))

(defun pig-extract-CiDian ()
 "*Convert the internal CiDian into the read-able CiDian table."
 (interactive)
 (let (filename)
  (save-excursion
   (set-buffer (get-buffer-create "Pig-CiDian-table"))
   (erase-buffer)
   (if Pig-in-Mule (set-file-coding-system *euc-china*))
   (mapcar 'pig-extract-hash-entry Pig-CD-hashed)
   (if (setq filename (expand-file-name
		       (read-file-name "Save the CiDian table to:"
				       Pig-CiDian-file)))
     (progn
      (write-file filename)
      (kill-buffer (buffer-name)))))))

(defun pig-extract-hash-entry (cd-entry)
 "Convert a whole entry of the hashed CiDian to it's ascii format."
 (mapcar 'pig-extract-entry (cdr cd-entry)))

(defun pig-extract-entry (entry)
 "Convert an CiDian entry to ascii format."
 (insert (mapconcat '(lambda (py) py) (car entry) "-")
	 "\t"
	 (mapconcat '(lambda (gb) gb) (car (cdr entry)) " ")
	 "\n"))

(defun pig-hot-key ()
 "*When you press 0-5, or space key, this command will start process the word
 right before this key."
 (interactive)
 (pig-word)
 (or Pig-8bit-capable (insert " ")))

(defun pig-instant-mode ()
 "*Make space a hot key."
 (interactive)
 (define-key Pig-mode-map " " 'pig-hot-key)
 (define-key Pig-mode-map "\C-c " '(lambda () 
				    (interactive) (insert-char ?  1)))
 (setq mode-name "PIG-instant")
 (redraw-display))

(defun pig-batch-mode ()
 "*Turn off hot number keys 0-5."
 (interactive)
 (define-key Pig-mode-map " " 'self-insert-command)
 (setq mode-name "PIG-batch")
 (redraw-display))

(defun pig-mode ()
 "Major mode for editing Chinese HanZi files using Pin1Yin1 input method.
   The best use is to type in PinYin with tones, and leave the converting 
   to GuoBiao at last with command \\[pig-buffer] or \\[pig-region] if only 
   part of the buffer needs this.
   Mule users can use \\[gb2hz-buffer] and \\[gb2hz-region] to convert all 
   HanZi to the HZ (7bit gb) coding.

   Special commands: (especially note those for special symbols)
   \\{Pig-mode-map}"
 (interactive)
 (if (boundp 'Pig-PinYin-table) () (pig-mode-initialization))
 (set-syntax-table Pig-mode-syntax-table)
 (if Pig-in-Mule (set-file-coding-system *euc-china*))
 (use-local-map Pig-mode-map)
 (setq mode-name "PIG-batch")
 (setq major-mode 'pig-mode))

(defun pig-dou-hao ()
 "*Insert GB for comma."
 (interactive)
 (pig-insert "\243\254"))

(defun pig-dun-hao ()
 "*Insert GB for dun-hao."
 (interactive)
 (pig-insert "\241\242"))

(defun pig-fen-hao ()
 "*Insert GB for semicolon."
 (interactive)
 (pig-insert "\243\273"))

(defun pig-gantan-hao ()
 "*Insert GB for exclamation mark."
 (interactive)
 (pig-insert "\243\241"))

(defun pig-mao-hao ()
 "*Insert GB for colon."
 (interactive)
 (pig-insert "\243\272"))

(defun pig-wen-hao ()
 "*Insert GB for question mark."
 (interactive)
 (pig-insert "\243\277"))

(defun pig-ju-hao ()
 "*Insert GB for Chinese period mark."
 (interactive)
 (pig-insert "\241\243"))

(defun pig-tilte ()
 "*Insert GB for tilte(~)."
 (interactive)
 (pig-insert "\241\253"))

(defun pig-left-bracket (double)
 "*Insert GB for left bracket, non-nil DOUBLE select double bracket <<."
 (interactive "P")
 (if double (pig-insert "\241\266")	; <<
  (pig-insert "\241\264")))	; <

(defun pig-right-bracket (double)
 "*Insert GB for right bracket, non-nil DOUBLE select double bracket >>."
 (interactive "P")
 (if double (pig-insert "\241\267")	; >>
  (pig-insert "\241\265")))	; >

;;;;;;;; Above are for both Emacs and Mule

(if Pig-8bit-capable  ; for 8bit Emacs and Mule
(progn

 (setq Pig-internal-prefix "")
 (if Pig-in-Mule (setq Pig-internal-prefix "\221"))

(defun pig-region (start end)
 "*Resolve all the PinYins in the region and replace them with the 
corrsponding gb codes. Mule or 8bit Emacs version."
 (interactive "r")
 (let
   ((current-syntax (syntax-table))
    (lend start)
    wb we word replaced)
  (unwind-protect
    (save-window-excursion
     (save-restriction
      (narrow-to-region start end)
      (goto-char (point-min))
      (set-syntax-table Pig-mode-syntax-table)
      (if (not (looking-at "\\<")) ;at start of word ?
	(re-search-forward "\\<" end 'stay)) ;start of word
      (while (re-search-forward "[ \\t]*\\(\\w+\\)" nil 'stay) ; next word
       (setq wb (match-beginning 1)
	     we (match-end 1)
	     word (downcase (buffer-substring wb we)))
       (if replaced 
	 (setq lend (match-beginning 0))
	(setq lend wb))
;;; replaced:   t, word is replaced, keep processing the new word
;;;           nil, skip this word
;;;          else, conversion to gb made, replace pinyin with gb
       (setq parsed (pig-py-parser word)
	     replaced (pig-choose parsed))
       (if (eq t replaced) ; replace the old pinyin
	 (progn
	  (setq word (read-string (format "Replacement for %s: " word)))
	  (delete-region wb we) ; replace pinyin
	  (insert-string word)
	  (goto-char lend))
	(if replaced
	  (progn
	   (delete-region lend we) ; replace, including blanks
	   (insert-string replaced))))) ; end of the last word processed
      (setq end (point))))
   (goto-char end)
   (set-syntax-table current-syntax)
   (message "Done."))))

(defun pig-add-cizu (py-list)
 "Given a phrase (list) of pinyin PY-LIST, get the exact spelling for each of
them and the corresponding gb. Return the new entry in CiDian.
If user abort any one character return nil. Mule and 8bit Emacs version."
 (let ((ci-chang (length py-list))
       (HZ-buffer (get-buffer-create " *HanZi*"))
       gb-list word plist gb-code (replaced t) (indx 0))
  (catch 'abort
   (while (< indx ci-chang)
    (setq word (nth indx py-list)
	  replaced t)
    (while (eq replaced t)
     (setq word (downcase word)
	   gb-list (pig-ZD-lookup word) ;a list of gb codes
	   replaced (pig-choose (list word gb-list nil))) ;get user choice
     (if (eq t replaced) ; replace the old pinyin
       (setq word (read-string (format "Replacement for %s: " word)))
      (if replaced			; if a selection has been made
	(setq plist (append plist (list word))
	      gb-code (concat gb-code replaced)
	      replaced nil)
       (throw 'abort nil))))
    (setq indx (1+ indx)))
   (save-excursion
    (set-buffer HZ-buffer)
    (erase-buffer)
    (insert-string gb-code)
    (setq mode-line-format (mapconcat '(lambda (string) string) plist "-")))
   (if (y-or-n-p "Is that correct ?")
     (list plist (list gb-code))
    t))))

(defun pig-word ()
 "*Find the pinyin word around or before the point, let user pick from a list
of possible gbs and replace the word with the choice. Mule and 8bit Emacs 
version."
 (interactive)
 (let
   ((current-syntax (syntax-table))
    wb we word parsed replaced)
  (unwind-protect
    (save-window-excursion
     (save-excursion
      (set-syntax-table Pig-mode-syntax-table)
      (if (not (looking-at "\\w"))
	(re-search-backward "\\w" (point-min) 'stay)) ;word end
      (re-search-backward "\\W" (point-min) 'stay) ; start of word
      (or (re-search-forward "\\w+" nil t)
	  (error "No Pin1Yin1 at or before the point."))
      (setq wb (match-beginning 0)
	    we (match-end 0)
	    word (buffer-substring wb we)
	    replaced t)
      (while (eq t replaced)
       (setq word (downcase word)
	     parsed (pig-py-parser word)
	     replaced (pig-choose parsed))
       (if (eq t replaced) ; replace the old pinyin
	 (progn
	  (setq word (read-string (format "Replacement for %s: " word)))
	  (delete-region wb we) ; replace pinyin
	  (insert-string word)
	  (setq we (point)))))
      (if replaced			; if a selection has been made
	(progn
	 (delete-region wb we)
	 (insert-string replaced)
	 (setq we (point))))))
   (if (< (point) we) (goto-char we))
   (set-syntax-table current-syntax))))
  
(defun pig-choose (parsed)
 "Take a list of PinYin, gb-list, CiZu indicator and the Zi list. Display
possible choices from the gb-list.  Return chosen word if one is chosen, or nil
to keep original word, or t for requesting replacement. Mule and 8bit Emacs
version."
 (unwind-protect 
   (save-window-excursion
    (let* ((word (nth 0 parsed))
	   (choices (nth 1 parsed))
	   (cizup (nth 2 parsed))
	   (count (length choices))
	   (indx 0)
	   (line 2)
	   (window-min-height 2)
	   (width (window-width))
	   (continue t)
	   (opt (if cizup ", a(dd)" ""))
	   (cur-buffer (current-buffer))
	   (HZ-buffer (get-buffer-create " *HanZi*"))
	   (HZ-window (get-buffer-window HZ-buffer))
	   new-entry help-msg next char num result)
     (save-excursion
      (set-buffer HZ-buffer)
      (erase-buffer)
      (buffer-flush-undo HZ-buffer)
      (setq truncate-lines t
	    mode-line-format 
	    (format "%s:: -- number to select, or r(etype), q(uit)%s --"
		    (upcase word) opt)
	    help-msg mode-line-format)
      (and Pig-in-Emacs (cxterm-mode)) ; for 8bit Emacs running in Cxterm
      (if (= count 0) 
	(progn
	 (insert "(none)")
	 (setq mode-line-format
	       (format "%s:: Nothing found -- Options: r(etype), q(uit)%s"
		       (upcase word) opt)
	       help-msg mode-line-format))
       (while (< indx count)
	(setq next (nth indx choices))
	(if (< (+ 6 (current-column) (length next)) width)
	  ()
	 (insert "\n")
	 (setq line (1+ line)))
	(insert "(" (+ indx ?0) ") " next "  ")
	(setq indx (1+ indx)))))
     (if HZ-window
       (progn
	(select-window HZ-window)
	(enlarge-window (- line (window-height))))
      (split-window-vertically line)
      (switch-to-buffer HZ-buffer))
     (select-window (get-buffer-window cur-buffer))
     (while continue
      (message help-msg)
      (setq char (read-char)
	    num (- char ?0))
      (cond ((= char ?r)                           ; Query replace
	     (setq result t
		   continue nil))
	    ((or (= char ? ) (= char ?q))
	     (setq result nil
		   continue nil))
	    ((and (>= num 0) (< num count))
	       (setq result (nth num choices)
		     continue nil))
	    ((and cizup (= char ?a))	; add a new entry to CiDian
	     (setq new-entry (pig-add-cizu (nth 3 parsed))
		   continue nil)
	     (if (eq new-entry t)
	       (setq result t)
	      (if new-entry
		(progn
		 (setq result (car (nth 1 new-entry)))
		 (pig-insert-cizu new-entry))
	       (setq result nil))))
	    ((or (= char help-char) (= char ?\?))
	     (message help-msg)
	     (sit-for 3))
	    (t (ding))))
     result))
  ;; Protected forms...
  (bury-buffer " *HanZi*")))

(defun pig-gb82d (gb)
 "Turn a GuoBiao string GB to display dependent string. Mule and 8bit Emacs
version."
 (let ((chang (/ (length gb) 2))
       gbd)
  (while (not (equal gb ""))
   (setq gbd (concat gbd Pig-internal-prefix (substring gb 0 2))
	 gb (substring gb 2)))
  gbd))

(if (fboundp 'gb2hz-buffer) ()
 (defun gb2hz-buffer ()
  "Convert whole text in the current buffer
from mule internal encoding to HZ encoding."
  (interactive) (gb2hz-region (point-min) (point-max))))

(defun pig-insert (gb)
 "*Insert GuoBiao for one character GB with proper convertion.
   In 7bit Emacs, strip the 8th bit
   In 8bit capable Emacs, simply insert the code
   In Mule, insert with prefix."
 (insert-string
  (concat Pig-internal-prefix gb)))

(if (fboundp 'gb2hz-region) ()		; We don't want to redefine these

(defun gb2hz-region (beg end)
  "Convert text in the current region
from mule internal encoding to HZ encoding."
  (interactive "r")
  (save-restriction
   (let (mc-flag)
    (narrow-to-region beg end)
    (goto-char (point-min))
    (replace-string "~" "~~")
    (goto-char (point-min))
    (replace-string "~~\n" "~\n")
    (code-convert-region (point-min) (point-max) *internal* *junet*)
    (goto-char (point-min))
    (replace-string "\e$A" "~{")
    (goto-char (point-min))
    (replace-string "\e(B" "~}")
    (goto-char (point-min))
    ))))

);;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; End of for Mule only

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; for Emacs only

;;; GB consists of two bytes. The highest bit of both bytes are on.
;;; The range of the higher byte is a1-f7 for standard Zi4Ku4 and
;;; that of the lower byte is a1-fe. The actual numbers, if used, here are the 
;;; GB bytes with their highest bit stripped, therefore they are in the range
;;; of ASCII.

(defvar GB7-Sections (list [33 47] [48 119] [120 121])
  "starting and ending code (of higher byte) for (7bit) GB symbols, GB hanzi
   and user defined hanzi.
   Lower byte always start from 33 and ends at 126.")

(defvar GB-b1-start 33)

(defvar GB-b1-end 126)

(defvar GB-b1-size 94
 "Number of characters with the same higher byte.")

(defvar Zi-Chang 32 
 "Number of bytes that one character has in the bitmap file ")

(defvar Zi-Kuan 16
 "Width of one character bitmap in bits.")

(defvar Zi-Gao 16
 "Height of characters in bits")

(defvar Pig-GBitmap-length nil
 "Number of characters in the GB font file.")

(defun pig-add-cizu (py-list)
 "Given a phrase (list) of pinyin PY-LIST, get the exact spelling for each of
them and the corresponding gb. Return the new entry in CIDian.
If user abort any one character return nil. 7bit Emacs version."
 (let ((ci-chang (length py-list))
       (HZ-buffer (get-buffer-create " *HanZi*"))
       (replaced t) (indx 0) gb-list word plist gb-code)
  (catch 'abort
   (while (< indx ci-chang)
    (setq word (nth indx py-list)
	  replaced t)
    (while (eq replaced t)
     (setq word (downcase word)
	   gb-list (pig-ZD-lookup word) ;a list of gb codes
	   replaced (pig-choose (list word gb-list nil))) ;get user choice
     (if (eq t replaced) ; replace the old pinyin
       (setq word (read-string (format "Replacement for %s: " word)))
      (if replaced			; if a selection has been made
	(setq plist (append plist (list word))
	      gb-code (concat gb-code replaced)
	      replaced nil)
       (throw 'abort nil))))
    (setq indx (1+ indx)))
   (save-excursion
    (set-buffer HZ-buffer)
    (erase-buffer)
    (pig-print-gb gb-code)
    (setq mode-line-format (mapconcat '(lambda (string) string) plist "-")))
   (if (y-or-n-p "Is that correct ?")
     (list plist (list (pig-gb728 gb-code)))
    t))))

(defun pig-region (start end)
 "*Resolve all the PinYins in the region and replace them with the 
corrsponding gb codes. 7bit Emacs version."
 (interactive "r")
 (let
   ((orig-buf (current-buffer))
    (current-syntax (syntax-table))
    (lend start)
    (HZ-buffer (get-buffer-create " *HanZi*"))
    wb we word parsed replaced skip)
  (unwind-protect
    (save-window-excursion
     (save-excursion
      (set-buffer HZ-buffer)
      (buffer-flush-undo HZ-buffer)
      (setq truncate-lines t)
      (erase-buffer))
     (split-window-vertically 17)
     (switch-to-buffer HZ-buffer) ; show HZ-buffer in the popup window
     (select-window (next-window))
     (set-syntax-table Pig-mode-syntax-table)
     (save-restriction
      (narrow-to-region start end)
      (goto-char start)
      (if (not (looking-at "\\<")) ;at start of word ?
	(re-search-forward "\\<" end 'stay)) ;start of word
      (while (re-search-forward "[ \\t]*\\(\\w+\\)" nil 'stay) ; next word
       (setq wb (match-beginning 1)
	     we (match-end 1)
	     word (downcase (buffer-substring wb we)))
       (if replaced 
	 (setq lend (match-beginning 0))
	(setq lend wb))
;;; replaced:   t, word is replaced, keep processing the new word
;;;           nil, skip this word
;;;          else, conversion to gb made, replace pinyin with gb
       (save-excursion
	(setq parsed (pig-py-parser word)
	      replaced (pig-choose parsed)))
       (if (eq t replaced) ; replace the old pinyin
	 (progn
	  (setq word (read-string (format "Replacement for %s: " word)))
	  (delete-region wb we) ; replace pinyin
	  (insert-string word)
	  (goto-char lend))
	(if replaced
	  (progn
	   (delete-region lend we) ; replace, including blanks
	   (insert-string replaced)
	   (setq lend (point)))
	 (setq skip t))))
      (setq end (point)))) ; end of the last word processed
   (bury-buffer HZ-buffer)
   (set-syntax-table current-syntax))
  (goto-char end)
  (if (and (not skip) (y-or-n-p "Review them ?"))
    (pig-hz-show-region start end)
   (message "Done."))))

(defun pig-word ()
 "*Find the pinyin word around or before the point, let user pick from a list
of possible gbs and replace the word with the choice. 7bit Emacs version."
 (interactive)
 (let
   ((current-syntax (syntax-table))
    (we (point))
    (HZ-buffer (get-buffer-create " *HanZi*"))
    wb word parsed replaced)
  (unwind-protect
    (save-window-excursion
     (save-excursion
      (set-syntax-table Pig-mode-syntax-table)
      (if (not (looking-at "\\w"))
	(re-search-backward "\\w" (point-min) 'stay)) ;word end
      (re-search-backward "\\W" (point-min) 'stay) ; start of word
      (or (re-search-forward "\\w+" we 'stay) ; Find start and end of word
	  (error "No Pin1Yin1 at or before the point."))
      (setq wb (match-beginning 0)
	    we (match-end 0)
	    word (buffer-substring wb we)
	    replaced t)
      (set-buffer HZ-buffer)
      (setq truncate-lines t)
      (buffer-flush-undo HZ-buffer)
      (erase-buffer)
      (split-window-vertically 17)
      (switch-to-buffer HZ-buffer) ; show HZ-buffer in the popup window
      (select-window (next-window))
      (while (eq t replaced)
       (setq word (downcase word)
	     parsed (pig-py-parser word)
	     replaced (pig-choose parsed))
       (if (eq t replaced) ; replace the old pinyin
	 (progn
	  (setq word (read-string (format "Replacement for %s: " word)))
	  (delete-region wb we)
	  (insert-string word)
	  (setq we (point)))))
     (if replaced			; if a selection has been made
       (progn
	(delete-region wb we)
	(insert replaced)
	(setq we (point))))))
   (if (< (point) we) (goto-char we))
   (bury-buffer HZ-buffer)
   (set-syntax-table current-syntax))))

(defun pig-choose (parsed)
 "Take a list of PinYin, gb-list, CiZu indicator and Zi list. Display possible
choices from the gb-list.  Return chosen word if one is chosen, or nil to keep
original word, or t for requesting replacement. 7bit Emacs version."
 (let*
   ((HZ-buffer (get-buffer-create " *HanZi*"))
    (HZ-window (get-buffer-window HZ-buffer))
    (word (nth 0 parsed))
    (choices (nth 1 parsed))
    (cizup (nth 2 parsed))
    (clist (pig-display-choices choices)) ; put everything in HanZi buffer
    (nscrn (1- (length clist)))
    (zi-shu (length choices))		;number of hanzi
    (before 0) (line 0)
    after start (count 0)
    (continue t)
    (opt (if cizup ", a(dd)" ""))
    char num result help-msg)
  (if clist 
    (setq count (nth 0 clist)
	  after (- zi-shu count)))
  (save-excursion
   (set-buffer HZ-buffer)
   (goto-char 0))
  (while continue			; not done yet
   (save-excursion
    (set-buffer HZ-buffer)
    (if (= 0 zi-shu)			; nothing found
      (setq mode-line-format 
	    (format "%s:: Nothing found -- Options: r(etype), q(uit)%s"
		    (upcase word) opt)
	    help-msg mode-line-format)
     (setq mode-line-format 
	   (format "%s:: %d<= %d =>%d -- [0-%d], r(etype), q(uit)%s ?(help) --"
		   (upcase word) before count after (1- count) opt)
	   result nil
	   help-msg (format "%s: Num to choose, > for more, < to go back" 
			    word (1- count))))
    (message help-msg))
   (setq char (read-char)	; get user input
	 num (- char ?0))
   (message " ")		;clear minibuffer
   (cond ((or (= char ? ) (= char ?>) (= char ?n)) ; ask to show more choices
	  (if (>= line nscrn)
	    (progn (beep 1) (message "No more"))
	   (setq line (1+ line)	; show more characters
		 before (+ before count)
		 count (nth line clist)
		 after (- zi-shu before count)
		 start (save-excursion
			(set-buffer " *HanZi*")
			(forward-line Zi-Gao) (point)))
	   (set-window-start HZ-window start)))
	 ((or (= char ?<) (= char ?p)) ;going back
	  (if (<= line 0)
	    (progn (beep 1) (message "No more"))
	   (setq line (1- line)
		 after (+ after count)
		 count (nth line clist)
		 before (- before count)
		 start (save-excursion
			(set-buffer " *HanZi*")
			(forward-line (- Zi-Gao)) (point)))
	   (set-window-start HZ-window start)))
	 ((and (>= num 0) (< num count)) ; a number in the range
	  (setq result (nth (+ num before) choices)
		continue nil))
	 ((= char ?r) ; retype pinyin, done in the caller
	  (setq result t
		continue nil))
	 ((and cizup (= char ?a))	; add a new entry to CiDian
	  (setq new-entry (pig-add-cizu (nth 3 parsed))
		continue nil)
	  (if (eq new-entry t)
	    (setq result t)
	   (if new-entry
	     (progn
	      (setq result (pig-gb82d (car (nth 1 new-entry))))
	      (pig-insert-cizu new-entry))
	    (setq result nil))))
	 ((= char ?q)			; quit without doing anything
	  (setq continue nil))
	 ((or (= char help-char) (= char ?\?))
	  (message help-msg)) ; do nothing
	 (t				; illegal 
	  (ding))))
  result))

(defun pig-display-choices (choices)
 "Display possible candidates from list CHOICES. Returns a list of 
number-of-selections for each line. For 7bit Emacs only."
 (let*
   ((ci-chang (if choices (/ (length (car choices)) 2) 1))
    (sel-per-line (max 1 (/ (window-width) (* ci-chang Zi-Kuan))))
    (zi-shu (length choices))		;number of hanzi
    (nscrn (/ zi-shu sel-per-line))	;number of screenful display
    (count (% zi-shu sel-per-line))
    (line 0) (seqn 0) (indx 0) clist)
  (save-excursion
   (set-buffer (get-buffer-create " *HanZi*"))
   (erase-buffer)
   (if (= 0 zi-shu)			; nothing to display
     (insert-string "(None)")
    (while (< line nscrn)
     (while (< indx sel-per-line)
      (pig-print-gb (nth seqn choices) (+ 48 indx))
      (setq indx (1+ indx)
	    seqn (1+ seqn)))
     (setq line (1+ line)
	   clist (append clist (list sel-per-line)))
     (setq indx 0)
     (goto-char (point-max))
     (insert "\n"))
    (if (zerop count) ()
     (while (< indx count)		; the last incomplete line
      (pig-print-gb (nth seqn choices) (+ 48 indx))
      (setq indx (1+ indx)
	    seqn (1+ seqn)))
     (setq clist (append clist (list count))))))
  clist))

(defun pig-view-buffer (&optional quiet)
 "*View all the HZ coded text in the current buffer. 7bit Emacs version."
 (interactive) 
 (pig-view-region (point-min) (point-max) quiet))	

(defun pig-view-region (beg end &optional quiet)
 "*View all the HZ coded text in the region, including English letters. 7bit
Emacs version."
 (interactive "r")
 (let
   ((cur-buf (current-buffer))
    (HZ-buffer (get-buffer-create " *HanZi*"))
    (start beg)
    (stop end)
    (hz-format nil))
  (unwind-protect
    (save-window-excursion
     (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))
      (replace-string "~~" "~")
      (goto-char (point-min))
      (or quiet (split-window-vertically 17))
      (switch-to-buffer HZ-buffer) ; show HZ-buffer in the popup window
      (setq HZ-window (selected-window))
      (buffer-flush-undo HZ-buffer)
      (erase-buffer)
      (setq truncate-lines t)
      (or quiet 
	  (progn
	   (setq mode-line-format "-- %b  Hit any key to continue.")
	   (select-window (next-window))))
      (switch-to-buffer cur-buf)
      (pig-break-view beg (point-max) quiet)))
   (bury-buffer HZ-buffer))))

(defun pig-break-view (beg end &optional quiet)
 "Break the region into HZ coded and pure English text and view them. 7bit
Emacs version."
 (if (re-search-forward "~{" end 'stay)
   (progn
    (setq hz-beg (point)
	  hz-format t)
    (pig-copy-region start (- hz-beg 2) quiet)
    (if (re-search-forward "~}" end 'stay)
      (setq start (point)
	    hz-end (- start 2))
     (setq hz-end (point)))
    (pig-hz-show hz-beg hz-end quiet))
  (if hz-format
    (pig-copy-region beg end quiet)
   (pig-hz-show beg end quiet)))
 (if (< (point) end) (pig-break-view start end quiet)))

(defun pig-copy-region (start end &optional quiet)
 "Copy all text in the region as is into HanZi buffer. 7bit Emacs version."
 (if (= start end) ()
  (let*
    ((region (buffer-substring start end))
     (HZ-buffer (get-buffer-create " *HanZi*"))
     (HZ-window (get-buffer-window HZ-buffer))
     column)
   (save-excursion
    (set-buffer HZ-buffer)
    (if (zerop (setq column (current-column)))
      ()
     (goto-char (point-max))
     (insert "\n")
     (set-window-start HZ-window (point)))
    (insert region))
   (or quiet (read-char)))))

(defun pig-hz-show (start end &optional quiet)
 "Show the macro-character of HanZi inthe region one frame at a time. For 7bit
Emacs only."
 (let
   ((current-syntax (syntax-table))
    (HZ-buffer (get-buffer-create " *HanZi*"))
    (hzw-wid (window-width))
    (gb7-buf (current-buffer))	; buffer contains 7bit GBs
    column  (ptr start)  HZ-window  word)
  (save-excursion
   (set-buffer HZ-buffer) ; show HZ-buffer in the popup window
   (setq HZ-window (get-buffer-window HZ-buffer)
	 column (current-column))
   (if (zerop column) ()
    (insert "\n")
    (setq column 0))
   (or quiet (set-window-start HZ-window (point)))
   (set-buffer gb7-buf)
   (set-syntax-table Pig-mode-syntax-table)
   (goto-char start)
   (while (< ptr end) ; still have gb to show
    (goto-char ptr)
    (if (not (looking-at "\\w"))	; word constituent ?
      (re-search-forward "\\W" end 'stay)) ; jump to next word
    (if (< end (setq ptr (+ 2 (point)))) ()
     (setq word (buffer-substring (- ptr 2) ptr)
	   column (+ column Zi-Kuan))
     (set-buffer HZ-buffer)
     (if (> column hzw-wid)	; need a form feed
       (progn
	(or quiet
	    (progn
	     (message "Hit any key to continue.")
	     (read-char)
	     (message " ")))	; clear the mess in mini-buffer
	(goto-char (point-max))
	(insert "\n")
	(or quiet (set-window-start HZ-window (point)))
	(setq column Zi-Kuan)))
     (pig-print-gb word ?&)
     (set-buffer gb7-buf)))
   (or quiet
       (progn
	(goto-char ptr)
	(message "Hit any key to continue.")
	(read-char)
	(message "That's all folks"))))))

(defun pig-hz-show-region (start end &optional quiet)
 "Show the macro-character of HanZi inthe region one frame at a time. 7bit
Emacs version."
 (let
   ((HZ-buffer (get-buffer-create " *HanZi*"))
    (gb7-buf (current-buffer)))	; buffer contains 7bit GBs
  (unwind-protect
    (save-window-excursion
     (split-window-vertically 17)
     (switch-to-buffer HZ-buffer) ; show HZ-buffer in the popup window
     (erase-buffer)
     (setq truncate-lines t)
     (or quiet
	 (setq mode-line-format "-- %b  Hit any key to continue."))
     (select-window (next-window))
     (set-buffer gb7-buf)
     (pig-hz-show start end quiet))
   (bury-buffer HZ-buffer))))

(defun pig-print-gb (gb &optional bit-mark column)
  "*Print out the bitmap of the character with GB starting from the current 
column in the window, using BIT-MARK for the bits. For 7bit Emacs only."
  (set-buffer " *HanZi*")
  (if (null column) (setq column (current-column)))
  (let*
    ((byte-per-line (/ Zi-Chang Zi-Gao))
     HZ-bitmap origin indx offset)
   (while (not (string-equal gb ""))
    (setq HZ-bitmap (pig-gb2bitmap gb)
	  byte-per-line (/ Zi-Chang Zi-Gao)
	  origin (point)
	  indx 0
	  offset nil)
    (if HZ-bitmap
      (progn
       (while (< indx (1- Zi-Chang))
	(insert-string (aref macro-bitmap (aref HZ-bitmap indx)))
	(setq indx (1+ indx))
	(if (= 0 (% indx byte-per-line)) 
	  (progn
	   (next-line 1)
	   (if (zerop (setq offset (- column (move-to-column column)))) ()
	    (insert-char ?  offset)))))
       (insert-string (aref macro-bitmap (aref HZ-bitmap indx)))
       (if bit-mark (subst-char-in-region (point-min) (point) ?& bit-mark t))
       (message " "))
     (insert gb))
    (goto-char origin)
    (end-of-line)
    (setq column (current-column))
    (setq gb (substring gb 2)))))

(defun pig-build-GB-bitmap ()
 "*Read-in the GB bitmap file into string GB-bitmap for later use. Only for
7bit Emacs."
 (save-excursion
  (get-buffer-create "GB-bitmap-buffer")
  (set-buffer "GB-bitmap-buffer")
  (if (boundp 'file-coding-system-for-read) 
    (setq file-coding-system-for-read *noconv*))
  (insert-file-contents Pig-font-file)
  (setq Pig-GBitmap-length (1- (point-max))
	GB-bitmap (buffer-string))
  (kill-buffer "GB-bitmap-buffer")))

(defun pig-gb2bitmap (gb)
 "Return the bitmap string for GB. For 7bit Emacs only."
 (let*
   ((gb-b1 (- (aref gb 0) (aref (nth 0 GB7-Sections) 0))) ;higher byte
    (gb-b2 (- (aref gb 1) GB-b1-start)) ;lower byte
    (index (* Zi-Chang (+ 1 gb-b2 (* GB-b1-size gb-b1)))))
  (if (< index Pig-GBitmap-length)
    (substring GB-bitmap (- index Zi-Chang) index))))

(defun pig-gb82d (gb)
 "Turn a GuoBiao string GB to display dependent string. Only for 7bit Emacs."
 (let ((chang (/ (length gb) 2))
       gbd)
  (while (not (equal gb ""))
   (setq gbd (concat gbd Pig-internal-prefix 
		     (char-to-string (logand 127 (aref gb 0)))
		     (char-to-string (logand 127 (aref gb 1))))
	 gb (substring gb 2)))
  gbd))

(defun pig-gb728 (gb)
 "Turn a 7bit GuoBiao string GB to 8bit GuoBiao string. Only for 7bit Emacs."
 (let ((chang (/ (length gb) 2))
       gbd)
  (while (not (equal gb ""))
   (setq gbd (concat gbd Pig-internal-prefix 
		     (char-to-string (logior 128 (aref gb 0)))
		     (char-to-string (logior 128 (aref gb 1))))
	 gb (substring gb 2)))
  gbd))

(defun pig-insert (gb)
 "*Insert GuoBiao for one character GB with proper convertion.
   In 7bit Emacs, strip the 8th bit
   In 8bit capable Emacs, simply insert the code
   In Mule, insert with prefix."
 (insert-string
  (concat (char-to-string (logand 127 (aref gb 0)))
	  (char-to-string (logand 127 (aref gb 1))))))

;;; this part for compatibility with Mule in Gnus only ;;;
(defun hz-add-hook (hook newfunc)
 (if (boundp hook)
   (let ((hook-value (symbol-value hook)))
    (if (and (listp hook-value)
	     (not (eq 'lambda (car hook-value))))
      (if (not (memq newfunc hook-value))
	(set hook (cons newfunc hook-value)))
     (set hook (list newfunc hook-value))))
  (set hook (list newfunc))))

(defun gnus-hz-group-p ()
 "Return non-nil if the current news group is a member of gnus-hz-group."
 (and gnus-newsgroup-name
      (pig-memequal gnus-newsgroup-name gnus-hz-group)))

(defvar gnus-Subject-mode-hook nil)
(defvar gnus-Article-mode-hook nil)
(defvar gnus-Article-prepare-hook nil)
(defvar gnus-Inews-article-hook nil)
(defvar gnus-hz-flag nil)
(defvar gnus-hz-group
 '("alt.chinese.text")
 "List of news groups in which Hz encoding is used.")

(hz-add-hook 'gnus-Subject-mode-hook 'gnus-Subject-mode-hook-hz)
(hz-add-hook 'gnus-Article-mode-hook 'gnus-Article-mode-hook-hz)
(hz-add-hook 'gnus-Article-prepare-hook 'gnus-Article-prepare-hook-hz)

(defun gnus-Subject-mode-hook-hz ()
 (define-key gnus-Subject-mode-map "z" 'gnus-toggle-hz))

(defun gnus-Article-mode-hook-hz ()
 (make-variable-buffer-local 'gnus-hz-flag))

(defun gnus-Article-prepare-hook-hz ()
 (if (gnus-hz-group-p)
   (setq gnus-hz-flag nil)))

(defun hz2gb-buffer ()
 "Convert whole text in the current buffer from HZ encoding to HanZi."
 (interactive)
 (pig-view-region (point-min) (point-max) t)
 (save-excursion
  (set-buffer gnus-Article-buffer)
  (let ((buffer-read-only nil))
   (erase-buffer)
   (insert-buffer " *HanZi*"))))

(defun gnus-toggle-hz ()
 (interactive)
 (if (gnus-hz-group-p)
   (let ((buffer (get-buffer gnus-Article-buffer)))
    (save-excursion
     (set-buffer gnus-Article-buffer)
     (let ((buffer-read-only nil))
      (buffer-flush-undo buffer)
      (if gnus-hz-flag 
	(progn
	 (setq truncate-lines nil)
	 (gnus-Subject-display-article gnus-current-article)
	 (gnus-Subject-beginning-of-article))
       (setq truncate-lines t)
       (hz2gb-buffer))
      (setq gnus-hz-flag (not gnus-hz-flag)))))))

;;; End of compatibility part ;;;;;

(setq macro-bitmap
      ["        " "       &" "      & " "      &&" "     &  " "     & &"
       "     && " "     &&&" "    &   " "    &  &" "    & & " "    & &&"
       "    &&  " "    && &" "    &&& " "    &&&&" "   &    " "   &   &"
       "   &  & " "   &  &&" "   & &  " "   & & &" "   & && " "   & &&&"
       "   &&   " "   &&  &" "   && & " "   && &&" "   &&&  " "   &&& &"
       "   &&&& " "   &&&&&"
       "  &     " "  &    &" "  &   & " "  &   &&" "  &  &  " "  &  & &"
       "  &  && " "  &  &&&" "  & &   " "  & &  &" "  & & & " "  & & &&"
       "  & &&  " "  & && &" "  & &&& " "  & &&&&" "  &&    " "  &&   &"
       "  &&  & " "  &&  &&" "  && &  " "  && & &" "  && && " "  && &&&"
       "  &&&   " "  &&&  &" "  &&& & " "  &&& &&" "  &&&&  " "  &&&& &"
       "  &&&&& " "  &&&&&&"
       " &      " " &     &" " &    & " " &    &&" " &   &  " " &   & &"
       " &   && " " &   &&&" " &  &   " " &  &  &" " &  & & " " &  & &&"
       " &  &&  " " &  && &" " &  &&& " " &  &&&&" " & &    " " & &   &"
       " & &  & " " & &  &&" " & & &  " " & & & &" " & & && " " & & &&&"
       " & &&   " " & &&  &" " & && & " " & && &&" " & &&&  " " & &&& &"
       " & &&&& " " & &&&&&"
       " &&     " " &&    &" " &&   & " " &&   &&" " &&  &  " " &&  & &"
       " &&  && " " &&  &&&" " && &   " " && &  &" " && & & " " && & &&"
       " && &&  " " && && &" " && &&& " " && &&&&" " &&&    " " &&&   &"
       " &&&  & " " &&&  &&" " &&& &  " " &&& & &" " &&& && " " &&& &&&"
       " &&&&   " " &&&&  &" " &&&& & " " &&&& &&" " &&&&&  " " &&&&& &"
       " &&&&&& " " &&&&&&&"
       "&       " "&      &" "&     & " "&     &&" "&    &  " "&    & &"
       "&    && " "&    &&&" "&   &   " "&   &  &" "&   & & " "&   & &&"
       "&   &&  " "&   && &" "&   &&& " "&   &&&&" "&  &    " "&  &   &"
       "&  &  & " "&  &  &&" "&  & &  " "&  & & &" "&  & && " "&  & &&&"
       "&  &&   " "&  &&  &" "&  && & " "&  && &&" "&  &&&  " "&  &&& &"
       "&  &&&& " "&  &&&&&"
       "& &     " "& &    &" "& &   & " "& &   &&" "& &  &  " "& &  & &"
       "& &  && " "& &  &&&" "& & &   " "& & &  &" "& & & & " "& & & &&"
       "& & &&  " "& & && &" "& & &&& " "& & &&&&" "& &&    " "& &&   &"
       "& &&  & " "& &&  &&" "& && &  " "& && & &" "& && && " "& && &&&"
       "& &&&   " "& &&&  &" "& &&& & " "& &&& &&" "& &&&&  " "& &&&& &"
       "& &&&&& " "& &&&&&&"
       "&&      " "&&     &" "&&    & " "&&    &&" "&&   &  " "&&   & &"
       "&&   && " "&&   &&&" "&&  &   " "&&  &  &" "&&  & & " "&&  & &&"
       "&&  &&  " "&&  && &" "&&  &&& " "&&  &&&&" "&& &    " "&& &   &"
       "&& &  & " "&& &  &&" "&& & &  " "&& & & &" "&& & && " "&& & &&&"
       "&& &&   " "&& &&  &" "&& && & " "&& && &&" "&& &&&  " "&& &&& &"
       "&& &&&& " "&& &&&&&"
       "&&&     " "&&&    &" "&&&   & " "&&&   &&" "&&&  &  " "&&&  & &"
       "&&&  && " "&&&  &&&" "&&& &   " "&&& &  &" "&&& & & " "&&& & &&"
       "&&& &&  " "&&& && &" "&&& &&& " "&&& &&&&" "&&&&    " "&&&&   &"
       "&&&&  & " "&&&&  &&" "&&&& &  " "&&&& & &" "&&&& && " "&&&& &&&"
       "&&&&&   " "&&&&&  &" "&&&&& & " "&&&&& &&" "&&&&&&  " "&&&&&& &"
       "&&&&&&& " "&&&&&&&&"])
);;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; End of for Emacs only

;;; Now do the initialization
(pig-mode-initialization)


;;Local variables:
;;lisp-body-indent: 1
;;end:
